import gc
import microcontroller
import ipaddress
import ssl
import socketpool
import adafruit_requests
import board
import adafruit_dotstar
import feathers2
import time
import digitalio
import adafruit_pct2075
import supervisor


# Calling garbage collector before importing large library.
#gc.collect()

# 2023-01-26 There seems to be a bug associated with the ESP-IDF code the 
# results in getting the error "MemoryError: Failed to allocate Wifi memory"
# when importing the wifi library. There is an open issue on adafruit associated 
# to this (https://github.com/adafruit/circuitpython/issues/7448). 
while True:
    try:
        import wifi
        break
    except (MemoryError) as error:
        print(f'Error: {error}')
        time.sleep(20)
        print('Hard resetting board...')
        microcontroller.reset()
        continue

# Uncomment next line when using Web Workflow and scanning for networks below.  
#wifi.radio.stop_scanning_networks()

# Create a DotStar instance
dotstar = adafruit_dotstar.DotStar(board.APA102_SCK, board.APA102_MOSI, 1, brightness=0.5, auto_write=True)

# Create I2C instance
i2c = board.I2C()

#bme280 = adafruit_bme280.Adafruit_BME280_I2C(i2c)
# The Feather may not detect the I2C from time to time. Therefore:
while True:
    try:
        #bme280 = adafruit_bme280.Adafruit_BME280_I2C(i2c)
        pct = adafruit_pct2075.PCT2075(i2c)
        break

    except (ValueError, RuntimeError) as error:
        print('Failed detecting I2C device. Retrying...')
        print(f'Error: {error}')
        time.sleep(5)
        print('Resetting board...')
        supervisor.reload()
        continue

# Creates a socketpool using the wifi radio's available sockets. 
# This is performed so we don't need to re-use sockets. 
pool = socketpool.SocketPool(wifi.radio)
requests = adafruit_requests.Session(pool, ssl.create_default_context())

# Constants
IO_POST_TIME = 60*20 # 20 min 
LOOP_INTERVAL = 60*2 # 2 min
TEMP_LIMIT = -15  #-15 Celsius. Depends on the Freezer temperature control.
WAITING_TIME = 60*70 # Waits this time before sending alarm when high temp. is detected
RETRY_TIMES = 5 # number of times to attempt to post
# LED colors
BLANK = (0, 0, 0)
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)


# Adafruit.IO and Homebridge configuration
HOMEBRIDGE_TEMP_ID = 'freezer-temp'
HOMEBRIDGE_TEMP_ALERT_ID = 'freezer-temp-alert'
IO_TEMP_FEED = 'freezer-temperature'
IO_ALERT_FEED = 'freezer-temperature-alert'

# Make sure the 2nd LDO is turned on
# feathers2.enable_LDO2(True)

# Turn on the internal blue LED
#feathers2.led_set(True)


def set_led(color=BLANK):
    dotstar[0] = color
    return

print("Freezer Temperature Sensor")

# ---- Wifi Connection ----
# Get wifi details and more from a secrets.py file
try:
    from secrets import secrets
except ImportError:
    print("WiFi secrets are kept in secrets.py, please add them there!")
    # LED turns Blue
    set_led(BLUE)
    raise

print("My MAC addr:", [hex(i) for i in wifi.radio.mac_address])
print(f'all> {wifi.radio.mac_address}')

# Uncomment below to scan and show available networks. 
# Note: If using Web Workflow, scanning for networks as done below may cause
# conflicts. Uncomment the line on top of the script 
# (wifi.radio.stop_scanning_networks()) to stop scanning for 
# networks from the beginning. 
#print("Available WiFi networks:")
#for network in wifi.radio.start_scanning_networks():
#    print("\t%s\t\tRSSI: %d\tChannel: %d" % (str(network.ssid, "utf-8"),
#            network.rssi, network.channel))
#wifi.radio.stop_scanning_networks()

print("Connecting to %s"%secrets["ssid"])
while True:
    try: 
        wifi.radio.connect(secrets["ssid"], secrets["password"])
        break
    except (ConnectionError) as error:
        print('Connection error. Retrying...')
        set_led(BLUE)
        time.sleep(0.5)
        continue
print("Connected to %s!"%secrets["ssid"])
print("My IP address is", wifi.radio.ipv4_address)

# ------------------------

def color_blink(color=(255,0,0)):

    for i in range(8):
        dotstar[0] = color
        time.sleep(0.2)
        dotstar[0] = (0, 0, 0)
        time.sleep(0.2)
    return


def post_adafruit(feed, data):

    # Tries several times to post to Adafruit. If it fails after so many times
    # it just continues. Adafruit.IO seems to have issues after a couple of 
    # days posting data to it.
    t = 1
    while t <= RETRY_TIMES:
        try:            
            print('Posting data to adafruit.io... ')
            
            payload = {"value": data}
            response = requests.post(
                "https://io.adafruit.com/api/v2/"
                + secrets["aio_username"]
                + "/feeds/"
                + feed
                + "/data",
                json=payload,
                headers={"X-AIO-KEY": secrets["aio_key"]},
            )
            print(f'Adafruit.io response: {response.json()}')
            response.close()
            print('OK')
            set_led()
            break

        except Exception as e:
            # Only time LED is set to green
            set_led(GREEN)
            print(f"Failed to post data to Adafruit.IO\nError: {e}\nTried {t} of 5 times.")
            time.sleep(10)
            t += 1
            continue
            
    # Response = None
    return

def post_boolean_homebridge(sensor, state):
    '''
    Sends data to Homebridge.
    Parameter:
    sensor: sensor ID as defined on the Homebridge configuration file. String.
    state: can be either 'true' or 'false'.
    '''

    # Tries several times to post. If it fails after so many times
    # it just continues. 
    t = 1
    while t <= RETRY_TIMES:
        try:
            print('Posting boolean data to Homebridge... ')
            
            payload = f"http://{secrets['server_ip']}/?accessoryId={sensor}&state={state}"
            response = requests.get(payload)
            print(f'Homebridge response: {response.json()}')
            response.close()
            print('OK')
            set_led()
            break

        #except (ValueError, RuntimeError) as e:
        except Exception as e:
            set_led(BLUE)
            print(f'Failed to post boolean to homebridge\nError: {e}')
            print(f'Tried {t} of {RETRY_TIMES} times.')
            time.sleep(10)
            t += 1
            continue

    #response = None
    return

def post_numeric_homebridge(sensor, value):
    '''
    Sends numeric value to Homebridge.
    Parameter:
    sensor: sensor ID as defined on the Homebridge configuration file. String.
    value: numeric value. Float or Int.
    '''

    # Tries several times to post. If it fails after so many times
    # it just continues. 
    t = 1
    while t <= RETRY_TIMES: 
        try:
            print('Posting numeric data to Homebridge... ')
            
            payload = f"http://{secrets['server_ip']}/?accessoryId={sensor}&value={value}"
            response = requests.get(payload)
            print(f'Homebridge response: {response.json()}')
            response.close()
            print('OK')
            set_led()
            break

        #except (ValueError, RuntimeError) as e:
        except Exception as e:
            set_led(BLUE)
            print(f'Failed to post numeric to homebridge\nError: {e}')
            print(f'Tried {t} of {RETRY_TIMES} times.')
            time.sleep(10)
            t += 1
            continue
            
    #response = None
    return


# Turn off LED after reset
set_led()
last_post = time.monotonic()
high_temp_time = time.monotonic()
alert_timer_started = False
alert_state_change = False
first_pass = True

while True:
            
    print('--------------------')

    while True:
        try:
            temperature_value = pct.temperature
            break

        except (OSError) as error:
            print('Failed reading temperature. Retrying...')
            print(f'Error: {error}')
            set_led(BLUE)
            time.sleep(2)
            continue

    
    if temperature_value > TEMP_LIMIT:
        high_temperature = True     
    else:
        high_temperature = False

    now = time.monotonic()

    # Send temperature information to Adafruit.IO
    # only at a predefined time interval (IO_POST_TIME).
    if (now - last_post >= IO_POST_TIME) or first_pass:

        last_post = now
        
        print('Posting temperature to Adafruit.IO... ')
        print(f'Temperature: {temperature_value} °C')
        post_adafruit(IO_TEMP_FEED, temperature_value)
        

    if high_temperature:

        if not alert_timer_started:
            alert_timer_started = True
            high_temp_time = now

        if now - high_temp_time >= WAITING_TIME:
            print(f'High temperature!.\nIt lasted more than the waiting time ({WAITING_TIME} secs.)')
            print('<< Sending Alerts >>')
            alert_state_change = True
            
            post_adafruit(IO_ALERT_FEED, 1)
            # Calling garbage collector after large functions
            #gc.collect()
            post_boolean_homebridge(HOMEBRIDGE_TEMP_ALERT_ID, 'false')
            #gc.collect()
            post_adafruit(IO_TEMP_FEED, temperature_value)
            #gc.collect()
            color_blink()
        
        else:
            print(f'High temperature!.\nIt has not lasted more than the waiting time ({WAITING_TIME} secs.)')
            print(f'{now - high_temp_time} seconds elapsed.')
    
    else:

        alert_timer_started = False

        # Only sends alert state to Adafruit if there is a change in its status.
        # Prevents constantly sending data to adafruit.io when alert is off.
        if alert_state_change or first_pass: 

            alert_state_change = False
            post_adafruit(IO_ALERT_FEED, 0)
            post_boolean_homebridge(HOMEBRIDGE_TEMP_ALERT_ID, 'true')
        
    # Posts numeric (temperature) data to Homebridge on every cycle.
    post_numeric_homebridge(HOMEBRIDGE_TEMP_ID, temperature_value)
    
    first_pass = False 

    time.sleep(LOOP_INTERVAL)
